/* Copyright (C) 1999-2005 Thomas S. Glascock */
/* www.soomka.com */

#include <pthread.h>
#include <getopt.h>
#include <string.h>
#include <ctype.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ncurses.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>

#define MINUTE 1
#define HOUR 2
#define SECOND 3
#define VERSION "0.2.2"

#define ALARMLENGTH (45)
#define SNOOZELENGTH (5)

char key;
int getkeys;

void *keygetcher(void *)
{
	extern char key;

	while (1)
	{
		pthread_testcancel();
		key=getch();
	}
}

inline int iseven(int numma)
{
	return (!(numma & 1));
}

inline float htod(int hours, int minutes)
{
	return	 	(
			((float)(hours>12?hours-12:hours) +
			((float)minutes / 60.0) // +
			) * 30.0 );
}

inline float mtod(int minutes, int seconds)
{
	return	 	(
			((float)minutes + 
			((float)seconds / 60.0) // +
			) * 6.0 );
}

inline float stod(int seconds, int microseconds)
{
	return	 	(
			((float)seconds +
			((float)microseconds  / 1000000.0)
			) * 6.0 );
}

class timex
{
	public:
		timex();
		~timex();
		void PrintFace();
		void PrintHands(float,float,float);
		void PrintHour(int,int);
		void PrintMinute(int,int);
		void PrintSecond(int,int);
		void RunClock();
		void Info();
		void Slide();
		void RedoWindows();
		void SaveSettings();
		void LoadSettings();

	private:
		int snooze;
		int silenced;
		int tripped;
		int alarmon;
		int alhour;
		int alminute;
		int alringing;
		int alarmlength;
		int snoozelength;
		int tmptimeout;
		int slider;
		int mainlines;
		int maincols;
		int slidelines;
		int slidecols;
		int slidex;
		int slidey;
		int slidexdir;
		int slideydir;
		int slidelock;
		long alarmend;
		long snoozeend;
		long tmpend;
		float pct[5];
		float pi;
		float aspect;
		float pisix;
		float pioneeighty;
		float yhalf;
		float xhalf;
		float ymult;
		float prevhour;
		float prevminute;
		float prevsecond;
		char *junk;
		char *symbol[5];
		WINDOW *facewin;
		WINDOW *handwin;
		WINDOW *hourwin;
		WINDOW *minutewin;
		WINDOW *secondwin;
};

timex::timex()
{
	pi=3.141592654;
	aspect=1.333333333;
	junk=(char *)calloc(1,256);

	symbol[1]=(char *)calloc(1,1);
	symbol[2]=(char *)calloc(1,1);
	symbol[3]=(char *)calloc(1,1);

	sprintf(symbol[1],"m");
	sprintf(symbol[2],"h");
	sprintf(symbol[3],"s");

	pct[HOUR]=60.0/100.0;
	pct[MINUTE]=95.0/100.0;
	pct[SECOND]=95.0/100.0;

	alarmon=0;
	alhour=0;
	alminute=0;
	alringing=0;
	snooze=0;
	silenced=0;
	tripped=0;
	alarmend=0;
	snoozeend=0;
	tmpend=0;
	tmptimeout=0;
	slider=0;
	slidexdir=1;
	slideydir=1;
	slidelock=0;

	alarmlength=ALARMLENGTH;
	snoozelength=SNOOZELENGTH;

	initscr();
	attrset(A_NORMAL);
	curs_set(0);
	halfdelay(0);
	noecho();
	mainlines=LINES;
	maincols=COLS;
	handwin=newwin(mainlines,maincols,0,0);
	facewin=newwin(mainlines,maincols,0,0); 
	hourwin=newwin(mainlines,maincols,0,0);
	minutewin=newwin(mainlines,maincols,0,0);
	secondwin=newwin(mainlines,maincols,0,0);
	prevhour=-1;
	prevminute=-1;
	prevsecond=-1;
}

timex::~timex()
{
	free(junk);
	curs_set(1);
	nocbreak();
	echo();
	nl();
	endwin();
}

void timex::SaveSettings()
{
	FILE *clockywockfile;
	char *filename;
	filename=(char *)calloc(1,256);
	strcat(filename,getenv("HOME"));
	strcat(filename,"/.clockywock");
	clockywockfile=fopen(filename,"w");
	fprintf(clockywockfile,"%d\n",alarmon);
	fprintf(clockywockfile,"%d\n",alhour);
	fprintf(clockywockfile,"%d\n",alminute);
	fprintf(clockywockfile,"%d\n",alringing);
	fprintf(clockywockfile,"%d\n",alarmlength);
	fprintf(clockywockfile,"%d\n",snoozelength);
	fprintf(clockywockfile,"%d\n",slider);
	fprintf(clockywockfile,"%c\n",*symbol[1]);
	fprintf(clockywockfile,"%c\n",*symbol[2]);
	fprintf(clockywockfile,"%c\n",*symbol[3]);
	fclose(clockywockfile);
	free(filename);
}

void timex::LoadSettings()
{
	FILE *clockywockfile;
	char *filename;
	filename=(char *)calloc(1,256);
	strcat(filename,getenv("HOME"));
	strcat(filename,"/.clockywock");
	clockywockfile=fopen(filename,"r");
	if (clockywockfile!=NULL)
	{
		fscanf(clockywockfile,(const char *)"%d\n",&alarmon);
		fscanf(clockywockfile,(const char *)"%d\n",&alhour);
		fscanf(clockywockfile,(const char *)"%d\n",&alminute);
		fscanf(clockywockfile,(const char *)"%d\n",&alringing);
		fscanf(clockywockfile,(const char *)"%d\n",&alarmlength);
		fscanf(clockywockfile,(const char *)"%d\n",&snoozelength);
		fscanf(clockywockfile,(const char *)"%d\n",&slider);
		fscanf(clockywockfile,(const char *)"%c\n",symbol[1]);
		fscanf(clockywockfile,(const char *)"%c\n",symbol[2]);
		fscanf(clockywockfile,(const char *)"%c\n",symbol[3]);
		fclose(clockywockfile);
	}
	free(filename);
}

void timex::Slide()
{
	if (slideydir==1)
	{
		if (slidey<slidelines)
			slidey=slidey+1;
		else
		{
			slideydir=-1;
		}
	}
	else
	{
		if (slidey>0)
			slidey=slidey-1;
		else
		{
			slideydir=1;
		}
	}

	if (slidexdir==1)
	{
		if (slidex<slidecols)
			slidex=slidex+1;
		else
		{
			slidexdir=-1;
		}
	}
	else
	{
		if (slidex>0)
			slidex=slidex-1;
		else
		{
			slidexdir=1;
		}
	}
	wclear(handwin);
	mvwin(facewin,slidey,slidex);
	mvwin(handwin,slidey,slidex);
	mvwin(hourwin,slidey,slidex);
	mvwin(minutewin,slidey,slidex);
	mvwin(secondwin,slidey,slidex);
}

void timex::Info()
{
	WINDOW *infowin;
	int y,x;
	int infolines,infocols;
	infolines=20;
	infocols=44;

	infowin=newwin(infolines,infocols,int(yhalf)-(infolines/2),int(xhalf)-(infocols/2));
	getmaxyx(infowin,y,x);
	box(infowin,0,0);

	sprintf(junk,"key       function                value");
	mvwaddstr(infowin,1,2,junk);
	sprintf(junk,"(1,!)     minute hand character:  \"%c\"",*symbol[1]);
	mvwaddstr(infowin,3,2,junk);
	sprintf(junk,"(2,@)     hour hand character:    \"%c\"",*symbol[2]);
	mvwaddstr(infowin,4,2,junk);
	sprintf(junk,"(3,#)     second hand character:  \"%c\"",*symbol[3]);
	mvwaddstr(infowin,5,2,junk);
	sprintf(junk,"(A)       alarm:                   %s",alarmon==1?"on":"off");
	mvwaddstr(infowin,7,2,junk);
	sprintf(junk,"(H,M)     alarm time:              %2.2d:%2.2d",alhour,alminute);
	mvwaddstr(infowin,9,2,junk);
	sprintf(junk,"(T)       alarm length:            %2.2dm",alarmlength);
	mvwaddstr(infowin,11,2,junk);
	sprintf(junk,"(Z)       snooze length:           %2.2dm",snoozelength);
	mvwaddstr(infowin,13,2,junk);
	sprintf(junk,"(S)       slider mode:             %s",slider==1?"on":"off");
	mvwaddstr(infowin,15,2,junk);
	sprintf(junk,"<space>   snoozzze                 %s",snooze==1?"on":"off");
	mvwaddstr(infowin,17,2,junk);
	sprintf(junk,"<enter>   kill alarm");
	mvwaddstr(infowin,18,2,junk);
	wrefresh(infowin);
	overwrite(infowin,handwin);
	delwin(infowin);
}

void timex::RedoWindows()
{
	wclear(handwin);
	wclear(facewin);
	wclear(hourwin);
	wclear(minutewin);
	wclear(secondwin);

	wrefresh(handwin);
	wrefresh(facewin);
	wrefresh(hourwin);
	wrefresh(minutewin);
	wrefresh(secondwin);

	delwin(handwin);
	delwin(facewin);
	delwin(hourwin);
	delwin(minutewin);
	delwin(secondwin);

	if (slider==1)
	{
		mainlines=(LINES-(LINES/8));
		maincols=(COLS-(COLS/8));
		slidelines=LINES-mainlines;
		slidecols=COLS-maincols;
		slidex=0;
		slidey=0;
	} else
	{
		mainlines=LINES;
		maincols=COLS;
	}

	xhalf=(maincols/2.0)+(iseven(maincols)?-0.5:0.0);
	yhalf=(mainlines/2.0)+(iseven(mainlines)?-0.5:0.0);
	pisix=pi/6.0;
	ymult=(yhalf/xhalf);
	pioneeighty=pi/180.0;

	handwin=newwin(mainlines,maincols,0,0);
	facewin=newwin(mainlines,maincols,0,0); 
	hourwin=newwin(mainlines,maincols,0,0);
	minutewin=newwin(mainlines,maincols,0,0);
	secondwin=newwin(mainlines,maincols,0,0);

	wrefresh(handwin);
	wrefresh(facewin);
	wrefresh(hourwin);
	wrefresh(minutewin);
	wrefresh(secondwin);

	prevhour=-1;
	prevminute=-1;
	prevsecond=-1;
}

void timex::PrintFace()
{
	int m;

	wclear(facewin);
	wattrset(facewin,A_STANDOUT);

	for (m=0;m<=11;++m)
	{
		mvwaddstr(facewin,
			(int)(yhalf + (  yhalf * (-cos(pisix*m)))),
			(int)(xhalf + (( xhalf * ( sin(pisix*m)))) / aspect),
			" ");
	}
	wattrset(facewin,A_NORMAL);
	sprintf(junk,"clockywock v.%s", VERSION);
	mvwaddstr(facewin,(int)(yhalf + (yhalf/3)),(int)(xhalf-(strlen(junk)/2)),junk); 
	mvwaddstr(facewin,(int)(yhalf),(int)(xhalf),"O");
}

void timex::PrintHour(int hour, int minute)
{
	float x;
	float sinz;
	float cosz;
	float handlength;
	float unit;

	werase(hourwin);

	unit=htod(hour,minute)*pioneeighty;

	cosz=((cos(unit))*ymult);
	sinz=((sin(unit))/aspect);;
	handlength=(int)(xhalf * pct[HOUR]);

	for (x=1;x!=handlength;++x)
	{
		mvwaddstr(hourwin, (int)(yhalf+(x*-cosz)), (int)(xhalf+(x*sinz)), symbol[HOUR]);
	}
	prevminute=minute;
}

void timex::PrintMinute(int minute, int second)
{
	float x;
	float sinz;
	float cosz;
	float handlength;
	float unit;

	werase(minutewin);

	unit=mtod(minute,second)*pioneeighty;

	cosz=((cos(unit))*ymult);
	sinz=((sin(unit))/aspect);;
	handlength=(int)(xhalf * pct[MINUTE]);
	for (x=1;x!=handlength;++x)
	{
		mvwaddstr(minutewin, (int)(yhalf+(x*-cosz)), (int)(xhalf+(x*sinz)), symbol[MINUTE]);
	}
	prevsecond=second;
}

void timex::PrintSecond(int second, int microsecond)
{
	float x;
	float sinz;
	float cosz;
	float handlength;
	float unit;

	unit=stod(second,microsecond)*pioneeighty;

	werase(secondwin);

	cosz=((cos(unit))*ymult);
	sinz=((sin(unit))/aspect);;
	handlength=(int)(xhalf * pct[SECOND]);
	for (x=1;x!=handlength;++x)
	{
		mvwaddstr(secondwin, (int)(yhalf+(x*-cosz)), (int)(xhalf+(x*sinz)), symbol[SECOND]);
	}
	prevsecond=second;
}


// void timex::PrintHands(float hourdeg, float minutedeg, float seconddeg)
// {
// 	float x;
// 	float sinz;
// 	float cosz;
// 	float handlength;
// 	float unit[4];
// 
// 	int increment;
// 
// 	overwrite(facewin,handwin);
// 
// 	unit[HOUR]=hourdeg*pioneeighty;
// 	unit[MINUTE]=minutedeg*pioneeighty;
// 	unit[SECOND]=seconddeg*pioneeighty;
// 
// 	attrset(A_NORMAL);
// 	
// 	for (increment=MINUTE;increment<=SECOND;++increment)
// 	{
// 		cosz=((cos(unit[increment]))*ymult);
// 		sinz=((sin(unit[increment]))/aspect);;
// 		handlength=(int)(xhalf * pct[increment]);
// 		for (x=1;x!=handlength;++x)
// 		{
// 			mvwaddstr(handwin, (int)(yhalf+(x*-cosz)), (int)(xhalf+(x*sinz)),
// 			symbol[increment]);
// 		}
// 	}
// 	overlay(facewin,handwin);
// 
// 	if (tmptimeout==1)
// 	{
// 		Info();
// 	}
// }

void timex::RunClock()
{
	extern char key;
	extern int getkeys;

	int prevlines=0;
	int prevcols=0;

	struct timeval td;
	struct tm *ct;

	pthread_t keythread;

	ct=(struct tm*)NULL;

	LoadSettings();

	if (getkeys == 1) pthread_create(&keythread,NULL,keygetcher,NULL);

	RedoWindows();

	for (;;)
	{
		if ((COLS!=prevcols)||(LINES!=prevlines))
		{
			pthread_cancel(keythread);
			RedoWindows();
			if (getkeys == 1) pthread_create(&keythread,NULL,keygetcher,NULL);
			PrintFace();
		}

		prevcols=COLS;
		prevlines=LINES;

		gettimeofday(&td,(struct timezone *)NULL);
		ct=localtime(&td.tv_sec);

		overwrite(facewin,handwin);

		/* if (ct->tm_min!=prevminute) PrintHour(ct->tm_hour,ct->tm_min); */
		PrintHour(ct->tm_hour,ct->tm_min);

		overlay(hourwin,handwin);

		/* if (( ((ct->tm_sec%15)==0) && (ct->tm_sec!=prevsecond) ) || (prevsecond==-1)) PrintMinute(ct->tm_min,ct->tm_sec); */

                PrintMinute(ct->tm_min,ct->tm_sec);

		overlay(minutewin,handwin);

		PrintSecond(ct->tm_sec,td.tv_usec);

		overlay(secondwin,handwin);

		overlay(facewin,handwin);

		if (tmptimeout==1)
		{
			Info();
		}

		if (alarmon==1)
		{
			if (silenced==0)
			{
				if (alarmend > td.tv_sec)
				{
					if (snoozeend < td.tv_sec)
					{
						alringing=1;
						snooze=0;
					} else
						alringing=0;
				} else
					alarmend=snooze=alringing=tripped=0;
			} else
			{
				if (alarmend < td.tv_sec)
					silenced=tripped=alringing=0;
				else
				{
					alringing=0;
				}
			}
			if (((ct->tm_hour==alhour)&&(ct->tm_min==alminute)))
			{
				if (tripped==0)
				{
					alarmend=td.tv_sec+(alarmlength * 60) ;
					tripped=1;
				}
			}
			if (alringing==1) { beep(); flash(); }
		}

		if (tmpend<td.tv_sec) { tmpend=tmptimeout=0; }

		// key is fetched in a separate thread

		if ((key!=-1))
		{
			tmpend=td.tv_sec+5;
			tmptimeout=1;

			switch (tolower(key))
			{
				case 'a':
					alarmon=(1-(alarmon)); break;
	
				case 'h':
					alhour = (alhour)<23 ? ++alhour : 0; break;
	
				case 'm':
					alminute = (alminute)<59 ? ++alminute : 0; break;
	
				case 't':
					alarmlength = (alarmlength)<60 ? ++alarmlength : 15; break;
	
				case 'z':
					snoozelength = (snoozelength)<30 ? ++snoozelength : 3; break;
	
				case '1':
					*symbol[1] = (*symbol[1])<126 ? ++*symbol[1] : 33; prevsecond=-1; break;
	
				case '!':
					*symbol[1] = (*symbol[1])>33 ? --*symbol[1] : 126; prevsecond=-1; break;
	
				case '2':
					*symbol[2] = (*symbol[2])<126 ? ++*symbol[2] : 33; prevminute=-1; break;
	
				case '@':
					*symbol[2] = (*symbol[2])>33 ? --*symbol[2] : 126; prevminute=-1; break;
	
				case '3':
					*symbol[3] = (*symbol[3])<126 ? ++*symbol[3] : 33; break;
	
				case '#':
					*symbol[3] = (*symbol[3])>33 ? --*symbol[3] : 126; break;

				case '/':
					tmpend=td.tv_sec; break;
	
				case 'q':
				{
					clear();
					wrefresh(handwin);
					SaveSettings();
					return;
					break;
				}
	
				case 's':
					wclear(handwin);
					wrefresh(handwin);
					slider=1-slider;
					pthread_cancel(keythread);
					RedoWindows();
					if (getkeys == 1) pthread_create(&keythread,NULL,keygetcher,NULL);
					PrintFace(); break;
	
				case ' ': 
					if ((alarmon==1) && (alringing==1))
					{
						snooze=1;
						snoozeend=td.tv_sec+ (snoozelength * 60);
					}
					break;
	
				case 10:
					if (alarmon==1) silenced=1; break;
	
				case 13:
					if (alarmon==1) silenced=1; break;
	
			}
			key=-1;
		}
		if ((slider==1)  && (slidelock==0) && (ct->tm_sec==0))
		{
			Slide();
			slidelock=1;
		} 

		if ((ct->tm_sec%5)==1) slidelock=0;

		if (slider==1) wrefresh(handwin);

		wrefresh(handwin);

		usleep(10000);
	}
}

int main(int argc, char **argv)
{
	extern char key;
	extern char* optarg;
	/* extern int optind, opterr, optopt; */
	int q;
	extern int getkeys;

	getkeys=1;

	key=-1;

	timex slimex;

	while (1) {
		q = getopt (argc, argv, "hk");
		printf ("q is [%d]\n",q);

		if (q == -1)
			break;

		switch (q) {
			case 'h':
				printf ("help\n");
				exit(0);
				break;
			case 'k':
				getkeys = 0;
				break;
		}
	}

	slimex.RunClock();
}
